class CircularQueue {
    _elements;
    _autoResize;
    _capacity;
    _size;
    _start;
    _end;
    constructor(props = {}) {
        const { capacity = 16, autoResize = true, autoResizeExpansionRatio = 1.5 } = props;
        if (!Number.isInteger(capacity) || capacity < 1) {
            throw new RangeError(`[CircularQueue] capacity is expected to be a positive integer, but got (${capacity}).`);
        }
        if (autoResize && autoResizeExpansionRatio < 1.2) {
            throw new RangeError(`[CircularQueue] autoResizeExpansionRatio is expected to be a number greater than 1.2, but got (${autoResizeExpansionRatio}).`);
        }
        this._capacity = capacity;
        this._size = 0;
        this._start = 0;
        this._end = -1;
        this._elements = new Array(capacity);
        this._autoResize = autoResize
            ? (nextExpectedSize) => {
                if (nextExpectedSize <= this._capacity)
                    return;
                const nextCapacity = Math.ceil(this._capacity * autoResizeExpansionRatio);
                this.resize(Math.max(nextCapacity, nextExpectedSize));
            }
            : () => { };
    }
    *[Symbol.iterator]() {
        const { _elements, _size, _start, _capacity } = this;
        if (_size == 0)
            return;
        const natureCount = _capacity - _start;
        if (_size <= natureCount) {
            for (let i = _start, _end = _start + _size; i < _end; ++i)
                yield _elements[i];
            return;
        }
        for (let i = _start, _end = _capacity; i < _end; ++i)
            yield _elements[i];
        for (let i = 0, _end = _size - natureCount; i < _end; ++i)
            yield _elements[i];
    }
    init(elements = [], start = 0, end = elements.length) {
        this.clear();
        if (start < 0)
            start = 0;
        if (end > elements.length)
            end = elements.length;
        if (start >= end)
            return;
        const count = end - start;
        if (count > this._capacity)
            this._capacity = count;
        const { _elements } = this;
        for (let i = 0, k = start; k < end; ++i, ++k)
            _elements[i] = elements[k];
        this._size = count;
        this._start = 0;
        this._end = count - 1;
    }
    resize(MAX_SIZE) {
        if (this._size > MAX_SIZE) {
            throw new RangeError(`[CircularQueue] failed to resize, the new queue space is insufficient.`);
        }
        this.rearrange();
        this._capacity = MAX_SIZE;
    }
    enqueue(element) {
        this._autoResize(this._size + 1);
        this._end = this._forwardIndex(this._end);
        this._elements[this._end] = element;
        if (this._size < this._capacity)
            this._size += 1;
        else
            this._start = this._forwardIndex(this._start);
    }
    unshift(element) {
        if (this._size === 0)
            this.enqueue(element);
        else {
            this._autoResize(this._size + 1);
            this._start = this._backwardIndex(this._start);
            this._elements[this._start] = element;
            if (this._size < this._capacity)
                this._size += 1;
            else
                this._end = this._backwardIndex(this._end);
        }
    }
    enqueues(elements, start = 0, end = elements.length) {
        if (start < 0)
            start = 0;
        if (end > elements.length)
            end = elements.length;
        if (start >= end)
            return;
        const cnt = end - start;
        const nextSize = this._size + cnt;
        this._autoResize(nextSize);
        const { _capacity, _elements } = this;
        if (nextSize < _capacity) {
            for (let k = start; k < end; k++) {
                this._end = this._forwardIndex(this._end);
                _elements[this._end] = elements[k];
            }
            this._size += cnt;
        }
        else {
            if (cnt < _capacity) {
                this._start = (this._start + nextSize - _capacity) % _capacity;
                for (let k = start; k < end; ++k) {
                    this._end = this._forwardIndex(this._end);
                    _elements[this._end] = elements[k];
                }
                this._size = _capacity;
            }
            else {
                for (let i = 0, k = start + cnt - _capacity; i < _capacity; ++i, ++k) {
                    _elements[i] = elements[k];
                }
                this._size = _capacity;
                this._start = 0;
                this._end = _capacity - 1;
            }
        }
    }
    dequeue(element) {
        if (this._size === 0) {
            if (element !== undefined) {
                this._elements[0] = element;
                this._size = 1;
                this._start = 0;
                this._end = 0;
            }
            return undefined;
        }
        const target = this._elements[this._start];
        this._start = this._forwardIndex(this._start);
        if (element === undefined)
            this._size -= 1;
        else {
            this._end = this._forwardIndex(this._end);
            this._elements[this._end] = element;
        }
        return target;
    }
    pop() {
        if (this._size === 0)
            return undefined;
        this._size -= 1;
        const target = this._elements[this._end];
        this._end = this._backwardIndex(this._end);
        return target;
    }
    splice(filter, elements, start, end) {
        let nextSize = 0;
        let nextEnd = this._start;
        const { _elements } = this;
        for (const element of this) {
            if (filter(element)) {
                _elements[nextEnd] = element;
                nextEnd = this._forwardIndex(nextEnd);
                nextSize += 1;
            }
        }
        this._size = nextSize;
        this._end = this._backwardIndex(nextEnd);
        if (elements !== undefined)
            this.enqueues(elements, start, end);
    }
    front() {
        return this._size === 0 ? undefined : this._elements[this._start];
    }
    back() {
        return this._size === 0 ? undefined : this._elements[this._end];
    }
    rearrange() {
        if (this._start === 0)
            return;
        const { _elements, _start, _size } = this;
        if (_size > 0) {
            if (_start + _size <= this._capacity) {
                for (let i = 0, k = _start; i < _size; ++i, ++k)
                    _elements[i] = _elements[k];
            }
            else {
                const tmpArray = [];
                for (const x of this)
                    tmpArray.push(x);
                for (let i = 0; i < _size; ++i)
                    _elements[i] = tmpArray[i];
                tmpArray.length = 0;
            }
        }
        this._start = 0;
        this._end = _size - 1;
    }
    destroy() {
        this._capacity = 0;
        this.clear();
        this._elements = null;
    }
    clear() {
        this._size = 0;
        this._start = 0;
        this._end = -1;
    }
    get size() {
        return this._size;
    }
    _forwardIndex(index) {
        const nextIndex = index + 1;
        return nextIndex === this._capacity ? 0 : nextIndex;
    }
    _backwardIndex(index) {
        const nextIndex = index - 1;
        return nextIndex < 0 ? this._capacity + nextIndex : nextIndex;
    }
}

class PriorityQueue {
    _compare;
    _elements = [];
    _size = 0;
    constructor(props) {
        this._compare = props.compare;
    }
    *[Symbol.iterator]() {
        const { _elements, _size } = this;
        for (let i = 1; i <= _size; ++i)
            yield _elements[i];
    }
    init(elements = [], start = 0, end = elements.length) {
        this.clear();
        if (start >= 0 && start < end && end <= elements.length) {
            const { _elements } = this;
            for (let i = start, k = this._size; i < end; ++i)
                _elements[++k] = elements[i];
            this._size = end - start;
            this._fastBuild();
        }
    }
    enqueue(element) {
        this._elements[++this._size] = element;
        this._up(this._size);
    }
    enqueues(elements, start = 0, end = elements.length) {
        if (start < 0)
            start = 0;
        if (end > elements.length)
            end = elements.length;
        if (start >= end)
            return;
        const N = end - start;
        if (N * 4 < this._size) {
            for (let i = start; i < end; ++i)
                this.enqueue(elements[i]);
            return;
        }
        const { _elements } = this;
        for (let i = start, k = this._size; i < end; ++i)
            _elements[++k] = elements[i];
        this._size += N;
        this._fastBuild();
    }
    dequeue(element) {
        if (this._size < 1) {
            if (element !== undefined) {
                this._size = 1;
                this._elements[1] = element;
            }
            return undefined;
        }
        const target = this._elements[1];
        if (element !== undefined) {
            this._elements[1] = element;
        }
        else {
            this._elements[1] = this._elements[this._size--];
        }
        this._down(1);
        return target;
    }
    splice(filter, elements = [], start = 0, end = elements.length) {
        const { _size, _elements } = this;
        let k = 0;
        for (let i = 1; i <= _size; ++i) {
            const element = _elements[i];
            if (filter(element))
                _elements[++k] = element;
        }
        if (start >= 0 && start < end && end <= elements.length) {
            for (let i = start; i < end; ++i)
                _elements[++k] = elements[i];
        }
        this._size = k;
        this._fastBuild();
    }
    front() {
        return this._size > 0 ? this._elements[1] : undefined;
    }
    destroy() {
        this.clear();
        this._elements = null;
    }
    clear() {
        this._size = 0;
    }
    get size() {
        return this._size;
    }
    _down(index) {
        const { _elements, _size, _compare } = this;
        for (let p = index, q; p <= _size; p = q) {
            const lft = p << 1;
            const rht = lft | 1;
            if (lft > _size)
                break;
            q = rht <= _size && _compare(_elements[rht], _elements[lft]) < 0 ? rht : lft;
            const tmp = _elements[q];
            if (_compare(_elements[p], tmp) <= 0)
                break;
            _elements[q] = _elements[p];
            _elements[p] = tmp;
        }
    }
    _up(index) {
        const { _elements, _compare } = this;
        for (let q = index, p; q > 1; q = p) {
            p = q >> 1;
            const tmp = _elements[q];
            if (_compare(_elements[p], tmp) <= 0)
                break;
            _elements[q] = _elements[p];
            _elements[p] = tmp;
        }
    }
    _fastBuild() {
        for (let p = this._size >> 1; p > 0; --p)
            this._down(p);
    }
}

export { CircularQueue, PriorityQueue };
